###############################################################################
#
# IBT: Isolated Build Tool
# Copyright (C) 2016, Richard Cook. All rights reserved.
#
# Simple wrappers around Docker etc. for fully isolated build environments
#
###############################################################################

from __future__ import print_function

from ibt.util import get_user_info

class Context(object):
    def __init__(self, working_dir):
        self._working_dir = working_dir
        self._user_info = get_user_info(working_dir)

    @property
    def working_dir(self): return self._working_dir

    def user_info(self): return self._user_info



###############################################################################
#
# IBT: Isolated Build Tool
# Copyright (C) 2017, Richard Cook. All rights reserved.
#
# Simple wrappers around Docker etc. for fully isolated build environments
#
###############################################################################

from __future__ import print_function
import os

from ibt.command import Command
from ibt.docker_util import docker_image_exists, docker_installed
from ibt.project import Project
from ibt.util import show_banner

def _show_system_info(ctx):
    uid, group_name, gid, user_name = ctx.user_info()

    print("System information:")
    print("  Docker:              {}".format("installed" if docker_installed() else "not installed"))
    print()

    print("Context information:")
    print("  Working directory:   {}".format(ctx.working_dir))
    print("  User:                {} ({})".format(user_name, uid))
    print("  Group:               {} ({})".format(group_name, gid))
    print()

def _show_project_info(project):
    project_uid, project_group_name, project_gid, project_user_name = project.user_info()

    print("Project information:")
    print("  Project directory:   {}".format(project.root_dir))
    print("  Project ID:          {}".format(project.project_id))
    print("  Configuration file:  {}".format(project.config_path))
    print("  Temporary directory: {}".format(project.dot_dir))
    print()

    print("Project user information:")
    print("  User:                {} ({})".format(project_user_name, project_uid))
    print("  Group:               {} ({})".format(project_group_name, project_gid))
    print()

    print("Docker container information:")
    print("  Docker image ID:     {}".format(project.image_id))
    container_project_dir = project.settings.get("container-project-dir", project.default_container_project_dir)
    print("  Project directory:   {}".format(container_project_dir))
    print("  Temporary directory: {}".format(project.container_dot_dir))
    docker = project.settings.get("docker", None)
    docker_image = None if docker is None else docker.get("image", None)
    if docker is None:
        print("  Docker base image:   not configured")
    else:
        print("  Docker base image:   {}".format(docker["image"]))
    print()

    print("IBT status:")
    print("  Temporary directory: {}".format("exists" if os.path.isdir(project.dot_dir) else "does not exist"))
    print("  Docker image:        {}".format("built" if docker_image_exists(project.image_id) else "not built"))

    volumes = project.settings.get("volumes", None)
    if volumes is not None and len(volumes) > 0:
        print("\nProject volumes:")
        for key in sorted(volumes):
            local_dir = project.resolve_local_path(key)
            print("  {} = {}".format(local_dir, volumes[key]))

    aliases = project.settings.get("aliases", None)
    if aliases is not None and len(aliases) > 0:
        print("\nProject aliases:")
        for key in sorted(aliases):
            alias = aliases[key]
            if isinstance(alias, list):
                print("  {}:".format(key))
                for line in alias:
                    print("  - {}".format(line))
            else:
                print("  {} = {}".format(key, alias))

class InfoCommand(Command):
    def __init__(self):
        super(InfoCommand, self).__init__("info", requires_project=False)

    def add_subparser(self, subparsers):
        p = subparsers.add_parser(self.name, help="Show configuration information")
        p.set_defaults(obj=self, handler=self.run)

    def run(self, ctx, _):
        locations = []
        project = Project.read(ctx.working_dir, locations)

        show_banner()
        _show_system_info(ctx)

        if project is None:
            print("No project found; looked in:")
            for location in locations:
                print("  {}".format(location))
            print()
        else:
            _show_project_info(project)



###############################################################################
#
# IBT: Isolated Build Tool
# Copyright (C) 2016, Richard Cook. All rights reserved.
#
# Simple wrappers around Docker etc. for fully isolated build environments
#
###############################################################################

from __future__ import print_function
import os
import re
import subprocess

from ibt.tracing import trace_command
from ibt.util import ensure_mount_sources

def _expand(env_vars, value):
    def _replace(m):
        key = m.group(1)
        replacement = env_vars.get(key)
        if replacement is None: replacement = os.getenv(key, None)
        return m.group(0) if replacement is None else replacement

    return re.sub('\$([A-Za-z_][A-Za-z_0-9]*)', _replace, value)

def _build_command(ctx, project, command_args, subcommand):
    container_project_dir = project.settings.get("container-project-dir", project.default_container_project_dir)

    rel_dir = os.path.relpath(ctx.working_dir, project.root_dir)
    container_working_dir = os.path.join(container_project_dir, rel_dir)

    additional_args = []

    ports = {}

    ports_setting = project.settings.get("ports", None)
    if ports_setting is not None:
        for key in ports_setting:
            host_port = key
            container_port = ports_setting[key]
            if host_port in ports:
                raise RuntimeError("Duplicate host port {}".format(host_port))
            ports[host_port] = container_port

    volumes = {
        project.root_dir: container_project_dir,
        project.dot_dir: project.container_dot_dir
    }
    _, _, _, user_name = ctx.user_info()

    env_vars = {
        "IBTPROJECTDIR": container_project_dir,
        "IBTUSER": user_name,
        "USER": user_name
    }

    env_vars_setting = project.settings.get("env_vars", None)
    if env_vars_setting is not None:
        for key in env_vars_setting:
            env_vars[key] = _expand(env_vars, env_vars_setting[key])

    volumes_setting = project.settings.get("volumes", None)
    if volumes_setting is not None:
        for key in volumes_setting:
            local_dir = _expand(env_vars, key)
            local_dir = project.resolve_local_path(local_dir)
            container_dir = volumes_setting[key]
            if local_dir in volumes:
                raise RuntimeError("Duplicate volume {}".format(local_dir))
            volumes[local_dir] = container_dir

    command = [
        "docker",
        "run",
        "--security-opt",
        "seccomp:unconfined",
        "-w",
        container_working_dir,
        "-u",
        user_name,
        "--rm"
    ] + \
        sum([["-p", "{}:{}".format(key, ports[key])] for key in ports], []) + \
        sum([["--volume", "{}:{}".format(key, _expand(env_vars, volumes[key]))] for key in volumes], []) + \
        sum([["--env", "{}={}".format(key, env_vars[key])] for key in env_vars], []) + \
        additional_args + \
        ([] if command_args is None else command_args) + \
        [project.image_id] + \
        ([] if subcommand is None else subcommand)

    return command, volumes

def check_process_in_container(ctx, project, args, command_args=None, subcommand=None):
    command, volumes = _build_command(ctx, project, command_args, subcommand)
    trace_command(args, command)
    with ensure_mount_sources(volumes.keys()):
        subprocess.check_call(command)

def call_process_in_container(ctx, project, args, command_args=None, subcommand=None):
    command, volumes = _build_command(ctx, project, command_args, subcommand)
    trace_command(args, command)
    with ensure_mount_sources(volumes.keys()):
        return subprocess.call(command)



###############################################################################
#
# IBT: Isolated Build Tool
# Copyright (C) 2016, Richard Cook. All rights reserved.
#
# Simple wrappers around Docker etc. for fully isolated build environments
#
###############################################################################

from __future__ import print_function
import os
import shutil

from ibt.command import Command
from ibt.docker_util import docker_image_exists
from ibt.util import temp_dir

class ScriptCommand(Command):
    def __init__(self):
        super(ScriptCommand, self).__init__("script", requires_project=True)

    def add_subparser(self, subparsers):
        p = subparsers.add_parser(self.name, help="Run script inside container")
        p.add_argument("script_path", metavar="SCRIPTPATH", help="script to run")
        p.set_defaults(obj=self, handler=self.run)

    def run(self, ctx, project, args):
        if not docker_image_exists(project.image_id):
            raise RuntimeError("Project has not been upped")

        with temp_dir(project.dot_dir) as dir:
            local_path = os.path.join(dir, "script")
            container_path = os.path.join(project.container_dot_dir, os.path.relpath(local_path, project.dot_dir))

            local_input_path = os.path.join(dir, "input")
            container_input_path = os.path.join(project.container_dot_dir, os.path.relpath(local_input_path, project.dot_dir))

            shutil.copyfile(args.script_path, local_input_path)

            with open(local_path, "wt") as f:
                f.write("#!/bin/sh\n")
                f.write("chmod +x {}\n".format(container_input_path))
                f.write(container_input_path + "\n")

            docker_run(ctx, args, container_path)



###############################################################################
#
# IBT: Isolated Build Tool
# Copyright (C) 2016, Richard Cook. All rights reserved.
#
# Simple wrappers around Docker etc. for fully isolated build environments
#
###############################################################################

from __future__ import print_function
import argparse

from ibt.command import Command
from ibt.project import Project
from ibt.util import get_commands, show_banner

# Yes, this is a hack
# I'm too lazy to refactor the command classes for now
def _get_command_help(command):
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()
    command.add_subparser(subparsers)
    return subparsers._get_subactions()[0].help

class HelpCommand(Command):
    def __init__(self):
        super(HelpCommand, self).__init__("help", requires_project=False)

    def add_subparser(self, subparsers):
        p = subparsers.add_parser(self.name, help="Display summary of standard commands")
        p.set_defaults(obj=self, handler=self.run)

    def run(self, ctx, args):
        show_banner()

        commands = get_commands()

        project = Project.read(ctx.working_dir)
        if project is None:
            has_aliases = False
        else:
            aliases = project.settings.get("aliases", None)
            has_aliases = aliases is not None and len(aliases) > 0

        max_name_len = 0
        for name in commands:
            if len(name) > max_name_len:
                max_name_len = len(name)

        if has_aliases:
            for key in aliases:
                if len(key) > max_name_len:
                    max_name_len = len(key)

        print("Commands:")
        for name in sorted(commands):
            command = commands[name]
            help = _get_command_help(command)
            print("  {}  {}".format(name.ljust(max_name_len), help))

        if has_aliases:
            print("\nProject aliases:")
            for key in sorted(aliases):
                alias = aliases[key]
                if isinstance(alias, list):
                    print("  {}:".format(key))
                    for line in alias:
                        print("  - {}".format(line))
                else:
                    print("  {}  {}".format(key.ljust(max_name_len), alias))

        print("")



###############################################################################
#
# IBT: Isolated Build Tool
# Copyright (C) 2016, Richard Cook. All rights reserved.
#
# Simple wrappers around Docker etc. for fully isolated build environments
#
###############################################################################

from __future__ import print_function

class Command(object):
    def __init__(self, name, requires_project):
        self._name = name
        self._requires_project = requires_project

    @property
    def name(self): return self._name

    @property
    def requires_project(self): return self._requires_project



###############################################################################
#
# IBT: Isolated Build Tool
# Copyright (C) 2017, Richard Cook. All rights reserved.
#
# Simple wrappers around Docker etc. for fully isolated build environments
#
###############################################################################

from __future__ import print_function
import hashlib
import os
import re
import sys
import yaml

from ibt.util import check_process, get_user_info

# Detects current project directory and configuration path based on following
# strategy:
# (1) Look in current directory for Ibtfile
# (2) Look in ancestor directories for Ibtfile up until root of file system
# (3) Look in user .ibtprojects for an entry corresponding to a parent of the
# current directory
class Project(object):
    @staticmethod
    def read(search_dir, locations=[]):
        def _find_matching_config_path(locations, projects, original_dir):
            user_dir = os.path.expanduser("~")
            longest_prefix = os.path.abspath("/")
            longest_dir = None
            longest_config_path = None
            for key in projects:
                full_dir = os.path.abspath(os.path.join(user_dir, key))
                locations.append(full_dir)
                prefix = os.path.commonprefix([full_dir, original_dir])
                if prefix == full_dir and len(prefix) > len(longest_prefix):
                    longest_prefix = prefix
                    longest_dir = full_dir
                    longest_config_path = os.path.abspath(os.path.join(user_dir, projects[key]))
            return longest_dir, longest_config_path

        def _read_config(locations, original_dir):
            user_project_path = os.path.expanduser("~/.ibtprojects")
            locations.append(user_project_path)
            if not os.path.isfile(user_project_path):
                return None

            with open(user_project_path, "rt") as f:
                config = yaml.load(f)

            projects = config.get("projects", None)
            if projects is None:
                return None

            root_dir, config_path = _find_matching_config_path(locations, projects, original_dir)
            return None if root_dir is None else Project(root_dir, config_path)

        def _read_helper(locations, original_dir, search_dir):
            config_path = os.path.join(search_dir, "Ibtfile")
            locations.append(config_path)
            if os.path.isfile(config_path):
                return Project(original_dir, config_path)

            parent_dir = os.path.dirname(search_dir)
            if parent_dir == search_dir:
                return _read_config(locations, original_dir)
            else:
                return _read_helper(locations, original_dir, parent_dir)

        return _read_helper(locations, search_dir, search_dir)

    def __init__(self, root_dir, config_path):
        self._root_dir = root_dir
        self._config_path = config_path
        self._project_id = hashlib.md5(self._root_dir).hexdigest()
        self._image_id = "ibt-{}".format(self._project_id)
        self._dot_dir = os.path.join(os.path.dirname(self._config_path), ".ibt")
        self._default_container_project_dir = "/{}".format(os.path.basename(self._root_dir))
        self._container_dot_dir = "/.ibt"
        self._settings = None
        self._user_info = get_user_info(self._root_dir)

    @property
    def root_dir(self): return self._root_dir

    @property
    def config_path(self): return self._config_path

    @property
    def project_id(self): return self._project_id

    @property
    def image_id(self): return self._image_id

    @property
    def dot_dir(self): return self._dot_dir

    @property
    def default_container_project_dir(self): return self._default_container_project_dir

    @property
    def container_dot_dir(self): return self._container_dot_dir

    @property
    def settings(self):
        if self._settings is None:
            with open(self._config_path, "rt") as f:
                self._settings = yaml.load(f)
                if self._settings is None:
                    raise RuntimeError("Invalid configuration file {}".format(self._config_path))
        return self._settings

    def resolve_local_path(self, path):
        return os.path.abspath(os.path.join(self._root_dir, path))

    def user_info(self): return self._user_info



###############################################################################
#
# IBT: Isolated Build Tool
# Copyright (C) 2017, Richard Cook. All rights reserved.
#
# Simple wrappers around Docker etc. for fully isolated build environments
#
###############################################################################

from __future__ import print_function
import colorama
import json

def _trace(s):
    print(colorama.Fore.YELLOW + s + colorama.Style.RESET_ALL)

def trace_command(args, command):
    if args.verbose:
        _trace(args, shell_join(command))

def trace_data(args, data):
    if args.verbose:
        s = json.dumps(data, indent=2)
        _trace(s)



#!/usr/bin/env python

###############################################################################
#
# IBT: Isolated Build Tool
# Copyright (C) 2016, Richard Cook. All rights reserved.
#
# Simple wrappers around Docker etc. for fully isolated build environments
#
###############################################################################

from __future__ import print_function
import argparse
import colorama
import os
import shlex
import sys

from ibt import __project_name__, __version__
from ibt.context import Context
from ibt.docker_util import docker_installed
from ibt.help_command import HelpCommand
from ibt.project import Project
from ibt.run_command import RunCommand
from ibt.util import get_commands

_HELP_COMMAND = HelpCommand()
_RUN_COMMAND = RunCommand()

class ThrowingArgumentParserError(Exception):
    def __init__(self, *args, **kwargs):
        super(ThrowingArgumentParserError, self).__init__(*args, **kwargs)

class ThrowingArgumentParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        super(ThrowingArgumentParser, self).__init__(*args, **kwargs)

    def error(self, message):
        raise ThrowingArgumentParserError(message)

    def parse_args_exit_on_error(self, argv):
        try:
            return self.parse_args(argv)
        except ThrowingArgumentParserError as e:
            self.print_usage(sys.stderr)
            self.exit(2, "{}: error: {}\n".format(self.prog, e.message))

    def parse_args_no_throw(self, argv):
        try:
            return self.parse_args(argv)
        except ThrowingArgumentParserError:
            return

def _format_alias_description(alias):
    if isinstance(alias, list):
        return "  Command:\n{}".format("\n".join(map(lambda x: "  $ {}".format(x), alias)))
    else:
        return "  Command:\n  $ {}".format(alias)

def _handle_alias(parser, project, ctx, args, alias):
    if isinstance(alias, list):
        _RUN_COMMAND.run_lines(ctx, project, args, alias)
    else:
        args = parser.parse_args(shlex.split(alias))
        args.handler(ctx, project, args)

def _main(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    colorama.init()
    working_dir = os.getcwd()
    ctx = Context(working_dir)

    parser = ThrowingArgumentParser(description="IBT: Isolated Build Tool (https://github.com/rcook/ibt)")
    parser.add_argument("--verbose", "-v", action="store_true", help="output diagnostic information including Docker commands executed by tool")
    parser.add_argument("--version", action="version", version="{} version {}".format(__project_name__, __version__))

    subparsers = parser.add_subparsers(
        title="Subcommands and aliases",
        description="Valid subcommands and aliases",
        help="Additional help")

    commands = get_commands()
    for key in sorted(commands):
        command = commands[key]
        command.add_subparser(subparsers)

    command_argv = argv

    # First handle commands that do not need a project
    args = parser.parse_args_no_throw(command_argv)
    if args is not None and not args.obj.requires_project:
        args.handler(ctx, args)
        return

    # Can't do much without Docker
    if not docker_installed():
        print("Please install and/or enable Docker")
        return

    # All commands from this point onwards require a project.
    # Furthermore, to get the full set of available commands, we need a project.
    project = Project.read(working_dir)
    if project is None:
        print("No Ibtfile project configuration file could be found")
        return

    # Special-case the "status" command to report status even
    # if .ibt directory has not been created yet
    if not os.path.isdir(project.dot_dir) and argv == ["status"]:
        StatusCommand().run(ctx, [])
        return

    aliases = project.settings.get("aliases", None)
    if aliases is not None:
        for key in sorted(aliases):
            alias = aliases[key]
            p = subparsers.add_parser(
                key,
                help="<alias>",
                description=_format_alias_description(alias),
                formatter_class=argparse.RawDescriptionHelpFormatter)
            p.set_defaults(handler=
                lambda ctx, project, args, alias=alias:
                    _handle_alias(parser, project, ctx, args, alias))

    args = parser.parse_args_exit_on_error(command_argv)
    args.handler(ctx, project, args)

if __name__ == "__main__":
    _main()



###############################################################################
#
# IBT: Isolated Build Tool
# Copyright (C) 2016, Richard Cook. All rights reserved.
#
# Simple wrappers around Docker etc. for fully isolated build environments
#
###############################################################################

from __future__ import print_function
import subprocess32

from ibt.container_util import check_process_in_container
from ibt.util import call_process, check_process

def docker_installed():
    try:
        return call_process(["docker", "--version"], timeout=1)
    except (OSError, subprocess32.TimeoutExpired):
        return False

def docker_image_exists(image_id):
    return call_process(["docker", "inspect", image_id])

def docker_image_build(image_id, context_dir):
    if docker_image_exists(image_id):
        print("Docker image {} already built".format(image_id))
    else:
        print("Building Docker image {}".format(image_id))
        check_process(["docker", "build", "-t", image_id, context_dir])

def docker_image_remove(image_id):
    if docker_image_exists(image_id):
        print("Destroying Docker image {}".format(image_id))
        check_process(["docker", "rmi", image_id])
    else:
        print("No Docker image {} to destroy".format(image_id))

def docker_run(ctx, project, args, container_run_path):
    check_process_in_container(ctx, project, args, None, ["/bin/sh", container_run_path])



###############################################################################
#
# IBT: Isolated Build Tool
# Copyright (C) 2016, Richard Cook. All rights reserved.
#
# Simple wrappers around Docker etc. for fully isolated build environments
#
###############################################################################

from __future__ import print_function

from ibt.command import Command
from ibt.docker_util import docker_image_remove

class DestroyCommand(Command):
    def __init__(self):
        super(DestroyCommand, self).__init__("destroy", requires_project=True)

    def add_subparser(self, subparsers):
        p = subparsers.add_parser(self.name, help="Destroy project image")
        p.set_defaults(obj=self, handler=self.run)

    def run(self, _0, project, _1):
        docker_image_remove(project.image_id)



###############################################################################
#
# IBT: Isolated Build Tool
# Copyright (C) 2016, Richard Cook. All rights reserved.
#
# Simple wrappers around Docker etc. for fully isolated build environments
#
###############################################################################

__project_name__ = "ibt"
__version__ = "0.5"
__description__ = "Simple wrappers around Docker etc. for fully isolated build environments"



###############################################################################
#
# IBT: Isolated Build Tool
# Copyright (C) 2016, Richard Cook. All rights reserved.
#
# Simple wrappers around Docker etc. for fully isolated build environments
#
###############################################################################

from __future__ import print_function
import argparse
import os

from ibt.command import Command
from ibt.docker_util import docker_image_exists, docker_run
from ibt.util import make_shell_script, temp_dir

class RunCommand(Command):
    def __init__(self):
        super(RunCommand, self).__init__("run", requires_project=True)

    def add_subparser(self, subparsers):
        p = subparsers.add_parser(self.name, help="Run command inside container")
        p.add_argument("command", metavar="COMMAND", help="command")
        p.add_argument("args", metavar="ARGS", nargs=argparse.REMAINDER, help="arguments")
        p.set_defaults(obj=self, handler=self.run)

    def run(self, ctx, project, args):
        if not docker_image_exists(project.image_id):
            raise RuntimeError("Project has not been upped")

        self.run_lines(ctx, project, args, [" ".join([args.command] + args.args)])

    def run_lines(self, ctx, project, args, lines):
        with temp_dir(project.dot_dir) as dir:
            local_path = os.path.join(dir, "script")
            container_path = os.path.join(project.container_dot_dir, os.path.relpath(local_path, project.dot_dir))
            make_shell_script(local_path, lines)
            docker_run(ctx, project, args, container_path)



#!/usr/bin/env python
###############################################################################
#
# IBT: Isolated Build Tool
# Copyright (C) 2016, Richard Cook. All rights reserved.
#
# Simple wrappers around Docker etc. for fully isolated build environments
#
###############################################################################

import os
import re

from setuptools import setup

_PROJECT_NAME = "ibt"

def _read_properties(project_name):
    init_path = os.path.abspath(os.path.join(project_name, "__init__.py"))
    regex = re.compile("^\\s*__(?P<key>.*)__\\s*=\\s*\"(?P<value>.*)\"\\s*$")
    with open(init_path, "rt") as f:
        props = {}
        for line in f.readlines():
            m = regex.match(line)
            if m is not None:
                props[m.group("key")] = m.group("value")

    return props

_PROPERTIES = _read_properties(_PROJECT_NAME)
project_name = _PROPERTIES["project_name"]
version = _PROPERTIES["version"]
description = _PROPERTIES["description"]

setup(
    name=project_name,
    version=version,
    description=description,
    setup_requires=["setuptools-markdown"],
    long_description_markdown_filename="README.md",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2.7",
    ],
    url="https://github.com/rcook/ibt",
    author="Richard Cook",
    author_email="rcook@rcook.org",
    license="MIT",
    packages=[project_name],
    install_requires=[
        "colorama",
        "pyyaml",
        "subprocess32"
    ],
    entry_points={
        "console_scripts": [
            "{0} = {0}.__main__:_main".format(project_name)
        ]
    },
    include_package_data=True,
    data_files=[("", ["README", "README.md"])],
    test_suite="{}.tests.suite".format(project_name),
    zip_safe=False)



###############################################################################
#
# IBT: Isolated Build Tool
# Copyright (C) 2016, Richard Cook. All rights reserved.
#
# Simple wrappers around Docker etc. for fully isolated build environments
#
###############################################################################

from __future__ import print_function
import argparse

from ibt.command import Command
from ibt.container_util import call_process_in_container
from ibt.docker_util import docker_image_exists

class ShellCommand(Command):
    def __init__(self):
        super(ShellCommand, self).__init__("shell", requires_project=True)

    def add_subparser(self, subparsers):
        p = subparsers.add_parser(self.name, help="Run interactive shell inside container")
        p.add_argument("command", metavar="COMMAND", nargs="?", help="command")
        p.add_argument("args", metavar="ARGS", nargs=argparse.REMAINDER, help="arguments")
        p.set_defaults(obj=self, handler=self.run)

    def run(self, ctx, project, args):
        if not docker_image_exists(project.image_id):
            raise RuntimeError("Project has not been upped")

        user_command = args.args if args.command is None else [args.command] + args.args
        status = call_process_in_container(ctx, project, args, ["-it"], user_command)
        if status != 0:
            print("Shell returned {}".format(status))



###############################################################################
#
# IBT: Isolated Build Tool
# Copyright (C) 2017, Richard Cook. All rights reserved.
#
# Simple wrappers around Docker etc. for fully isolated build environments
#
###############################################################################

from __future__ import print_function
import contextlib
import grp
import os
import pwd
import re
import shutil
import subprocess32 as subprocess
import tempfile

def _flatten(*args):
    output = []
    for arg in args:
        if hasattr(arg, "__iter__"):
            output.extend(_flatten(*arg))
        else:
            output.append(arg)
    return output

def get_commands():
    if get_commands.commands is None:
        from ibt.destroy_command import DestroyCommand
        from ibt.help_command import HelpCommand
        from ibt.info_command import InfoCommand
        from ibt.run_command import RunCommand
        from ibt.script_command import ScriptCommand
        from ibt.shell_command import ShellCommand
        from ibt.up_command import UpCommand

        commands = [
            DestroyCommand(),
            HelpCommand(),
            InfoCommand(),
            RunCommand(),
            ScriptCommand(),
            ShellCommand(),
            UpCommand()
        ]
        get_commands.commands = {}
        for command in commands:
            get_commands.commands[command.name] = command

    return get_commands.commands
get_commands.commands = None

def call_process(command, timeout=None):
    proc = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE)
    proc.communicate(timeout=timeout)
    return proc.returncode == 0

def check_process(command):
    proc = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE)
    (out, error) = proc.communicate()
    if proc.returncode != 0:
        raise RuntimeError("Popen command failed: {} (command: {})".format(error, command))
    return out

def make_shell_script(path, lines):
    with open(path, "wt") as f:
        f.write("#!/bin/sh\n")
        for line in lines:
            f.write(line + "\n")

@contextlib.contextmanager
def temp_file(dir=None):
    f, temp_path = tempfile.mkstemp(dir=dir)
    os.close(f)
    try:
        yield temp_path
    finally:
        if os.path.isfile(temp_path):
            os.unlink(temp_path)

@contextlib.contextmanager
def ensure_mount_sources(*paths):
    try:
        cleanup_dirs = []
        for p in _flatten(paths):
            if p is not None and not os.path.isdir(p) and not os.path.isfile(p):
                cleanup_dirs.append(p)
                os.makedirs(p)
        yield
    finally:
        for dir in cleanup_dirs:
            try:
                os.removedirs(dir)
            except OSError:
                pass

@contextlib.contextmanager
def temp_dir(dir=None):
    with ensure_mount_sources(dir):
        temp_path = tempfile.mkdtemp(dir=dir)
        try:
            yield temp_path
        finally:
            if os.path.isdir(temp_path):
                shutil.rmtree(temp_path)

def shell_join(command):
    return " ".join(pipes.quote(s) for s in command)

def show_banner():
    print("IBT: Isolated Build Tool")
    print("https://github.com/rcook/ibt")
    print()

def get_user_info(working_dir):
    def _sanitize(s):
        return re.sub("[{}]".format(re.escape("^")), "_", s)

    fs = os.stat(working_dir)
    uid = fs.st_uid
    uname = pwd.getpwuid(uid)[0]
    gid = fs.st_gid
    gname = grp.getgrgid(gid)[0]

    # Note the weird order we return the information in...
    return [str(uid), _sanitize(gname), str(gid), _sanitize(uname)]



###############################################################################
#
# IBT: Isolated Build Tool
# Copyright (C) 2016, Richard Cook. All rights reserved.
#
# Simple wrappers around Docker etc. for fully isolated build environments
#
###############################################################################

from __future__ import print_function
import os
import subprocess
import pipes

from ibt.command import Command
from ibt.docker_util import docker_image_build, docker_image_remove
from ibt.tracing import trace_data
from ibt.util import make_shell_script, temp_dir, temp_file

class UpCommand(Command):
    def __init__(self):
        super(UpCommand, self).__init__("up", requires_project=True)

    def add_subparser(self, subparsers):
        p = subparsers.add_parser(self.name, help="Create project image")
        p.add_argument("--destroy", "-d", action="store_true", help="destroy before recreating project image")
        p.add_argument("--docker-build", "-b", action="store_true", help="rebuild base image before creating project image")
        p.add_argument("--config", "-c", help="specify a Docker configuration")
        p.set_defaults(obj=self, handler=self.run)

    def run(self, ctx, project, args):
        if args.config is None:
            parent_settings = project.settings
        else:
            configs = project.settings["configurations"]
            parent_settings = configs.get(args.config)
            if parent_settings is None:
                raise RuntimeError("Configuration {} not found (available configurations: {})".format(args.config, ", ".join(configs.keys())))

        docker = parent_settings.get("docker", None)
        if docker is None:
            docker_image = None
            docker_build = None
        else:
            docker_image = docker.get("image", None)
            docker_build = docker.get("build", None)

        trace_data(args, docker)

        if args.docker_build:
            if docker_build is None:
                raise RuntimeError("No build commands configured for Docker base image")
            else:
                with temp_file() as temp_path:
                    make_shell_script(temp_path, ["cd {}".format(project.root_dir)] + docker_build)
                    subprocess.check_call(["/bin/sh", temp_path])

        if args.destroy:
            docker_image_remove(project.image_id)

        with temp_dir() as dir:
            if docker_image is None:
                raise RuntimeError("No Docker base image is configured")
            else:
                uid, group_name, gid, user_name = ctx.user_info()
                with open(os.path.join(dir, "Dockerfile"), "wt") as f:
                    f.write("FROM {}\n".format(docker_image))
                    f.write("RUN groupadd -g {} {}\n".format(gid, pipes.quote(group_name)))
                    f.write("RUN useradd -l -u {} -g {} {}\n".format(uid, gid, pipes.quote(user_name)))

            with open(os.path.join(dir, ".dockerignore"), "wt") as f:
                f.write("*\n")

            docker_image_build(project.image_id, dir)
